@@ -96,7 +96,14 @@ extern int yydebug;
     TOK_SSLv3 = 306,
     TOK_TLSv1_0 = 307,
     TOK_TLSv1_1 = 308,
-    TOK_TLSv1_2 = 309
+    TOK_TLSv1_2 = 309,
+    TOK_SESSION_CACHE = 310,
+    TOK_SHARED_CACHE_LISTEN = 311,
+    TOK_SHARED_CACHE_PEER = 312,
+    TOK_SHARED_CACHE_IF = 313,
+    TOK_PRIVATE_KEY = 314,
+    TOK_BACKEND_REFRESH = 315,
+    TOK_OCSP_REFRESH_INTERVAL = 316
   };
 #endif
 /* Tokens.  */
@@ -152,18 +159,25 @@ extern int yydebug;
 #define TOK_TLSv1_0 307
 #define TOK_TLSv1_1 308
 #define TOK_TLSv1_2 309
+#define TOK_SESSION_CACHE 310
+#define TOK_SHARED_CACHE_LISTEN 311
+#define TOK_SHARED_CACHE_PEER 312
+#define TOK_SHARED_CACHE_IF 313
+#define TOK_PRIVATE_KEY 314
+#define TOK_BACKEND_REFRESH 315
+#define TOK_OCSP_REFRESH_INTERVAL 316
 
 /* Value type.  */
 #if ! defined YYSTYPE && ! defined YYSTYPE_IS_DECLARED
 
 union YYSTYPE
 {
-#line 31 "cfg_parser.y" /* yacc.c:1909  */
+#line 36 "cfg_parser.y" /* yacc.c:1909  */
 
 	int	i;
 	char	*s;
 
-#line 167 "cfg_parser.h" /* yacc.c:1909  */
+#line 181 "cfg_parser.h" /* yacc.c:1909  */
 };
 
 typedef union YYSTYPE YYSTYPE;
