@@ -4,22 +4,26 @@
  * Original author: Brane F. Gracnar
  */
 
+#include "config.h"
+
+#include <sys/types.h>
+#include <sys/stat.h>
+
 #include <stdio.h>
 #include <string.h>
+#include <strings.h>
 #include <errno.h>
 #include <stdarg.h>
 #include <stdlib.h>
 #include <ctype.h>
 #include <unistd.h>
 #include <getopt.h>
-#include <sys/types.h>
 #include <pwd.h>
 #include <grp.h>
-#include <sys/stat.h>
 #include <syslog.h>
 #include <libgen.h>
+#include <limits.h>
 
-#include "config.h"
 #include "configuration.h"
 #include "foreign/miniobj.h"
 #include "foreign/vas.h"
@@ -41,6 +45,7 @@
 #define CFG_WORKERS "workers"
 #define CFG_BACKLOG "backlog"
 #define CFG_KEEPALIVE "keepalive"
+#define CFG_BACKEND_REFRESH "backendrefresh"
 #define CFG_CHROOT "chroot"
 #define CFG_USER "user"
 #define CFG_GROUP "group"
@@ -48,6 +53,8 @@
 #define CFG_SYSLOG "syslog"
 #define CFG_SYSLOG_FACILITY "syslog-facility"
 #define CFG_PARAM_SYSLOG_FACILITY 11015
+#define CFG_PARAM_SEND_BUFSIZE 11016
+#define CFG_PARAM_RECV_BUFSIZE 11017
 #define CFG_DAEMON "daemon"
 #define CFG_WRITE_IP "write-ip"
 #define CFG_WRITE_PROXY "write-proxy"
@@ -95,11 +102,11 @@ void cfg_cert_file_free(struct cfg_cert_file **cfptr);
 static char error_buf[CONFIG_BUF_SIZE];
 static char tmp_buf[150];
 
-/* declare static printf like functions: */
-static void config_error_set(char *fmt, ...)
+/* declare printf like functions: */
+void config_error_set(char *fmt, ...)
 	__attribute__((format(printf, 1, 2)));
 
-static void
+void
 config_error_set(char *fmt, ...)
 {
 	int len;
@@ -116,7 +123,7 @@ config_error_set(char *fmt, ...)
 	memcpy(error_buf, buf, len);
 }
 
-char *
+const char *
 config_error_get(void)
 {
 	return error_buf;
@@ -158,12 +165,14 @@ front_arg_destroy(struct front_arg *fa)
 hitch_config *
 config_new(void)
 {
+	int i;
 	hitch_config *r;
 	struct front_arg *fa;
 
 	r = calloc(1, sizeof(hitch_config));
 	AN(r);
 
+	(void) i;
 	// set default values
 
 	r->PMODE              = SSL_SERVER;
@@ -197,13 +206,14 @@ config_new(void)
 	r->OCSP_VFY = 0;
 	r->OCSP_RESP_TMO = 10.0;
 	r->OCSP_CONN_TMO = 4.0;
+	r->OCSP_REFRESH_INTERVAL = 1800;
 
 #ifdef USE_SHARED_CACHE
 	r->SHARED_CACHE       = 0;
 	r->SHCUPD_IP          = NULL;
 	r->SHCUPD_PORT        = NULL;
 
-	for (int i = 0 ; i < MAX_SHCUPD_PEERS; i++)
+	for (i = 0 ; i < MAX_SHCUPD_PEERS; i++)
 		memset(&r->SHCUPD_PEERS[i], 0, sizeof(shcupd_peer_opt));
 
 	r->SHCUPD_MCASTIF     = NULL;
@@ -214,6 +224,7 @@ config_new(void)
 	r->SYSLOG             = 0;
 	r->SYSLOG_FACILITY    = LOG_DAEMON;
 	r->TCP_KEEPALIVE_TIME = 3600;
+	r->BACKEND_REFRESH_TIME = 0;
 	r->DAEMONIZE          = 0;
 	r->PREFER_SERVER_CIPHERS = 0;
 	r->TEST	              = 0;
@@ -230,7 +241,7 @@ config_new(void)
 	r->RING_SLOTS = 0;
 	r->RING_DATA_LEN = 0;
 
-	return r;
+	return (r);
 }
 
 void
@@ -257,9 +268,8 @@ config_destroy(hitch_config *cfg)
 		cfg_cert_file_free(&cf);
 	}
 
-	if (cfg->CERT_DEFAULT != NULL) {
+	if (cfg->CERT_DEFAULT != NULL)
 		cfg_cert_file_free(&cfg->CERT_DEFAULT);
-	}
 
 	free(cfg->CIPHER_SUITE);
 	free(cfg->ENGINE);
@@ -269,10 +279,11 @@ config_destroy(hitch_config *cfg)
 	free(cfg->ALPN_PROTOS_LV);
 
 #ifdef USE_SHARED_CACHE
+	int i;
 	free(cfg->SHCUPD_IP);
 	free(cfg->SHCUPD_PORT);
 
-	for (int i = 0; i < MAX_SHCUPD_PEERS; i++) {
+	for (i = 0; i < MAX_SHCUPD_PEERS; i++) {
 		free(cfg->SHCUPD_PEERS[i].ip);
 		free(cfg->SHCUPD_PEERS[i].port);
 	}
@@ -283,20 +294,21 @@ config_destroy(hitch_config *cfg)
 	free(cfg);
 }
 
-char *
+static char *
 config_assign_str(char **dst, char *v)
 {
 	assert(v != NULL);
 
-	if (strlen(v) <= 0) return(NULL);
+	if (strlen(v) <= 0)
+		return(NULL);
 	if (*dst != NULL)
 		free(*dst);
 
 	*dst = strdup(v);
-	return *dst;
+	return (*dst);
 }
 
-int
+static int
 config_param_val_bool(char *val, int *res)
 {
 	assert(val != NULL);
@@ -307,10 +319,10 @@ config_param_val_bool(char *val, int *res)
 		*res = 1;
 	}
 
-	return 1;
+	return (1);
 }
 
-int
+static int
 config_param_host_port_wildcard(const char *str, char **addr,
     char **port, char **cert, int wildcard_okay)
 {
@@ -318,12 +330,12 @@ config_param_host_port_wildcard(const char *str, char **addr,
 
 	if (str == NULL) {
 		config_error_set("Invalid/unset host/port string.");
-		return 0;
+		return (0);
 	}
 
 	if (strlen(str) > ADDR_LEN) {
 		config_error_set("Host address too long.");
-		return 0;
+		return (0);
 	}
 
 	// address/port buffers
@@ -336,27 +348,28 @@ config_param_host_port_wildcard(const char *str, char **addr,
 	// FORMAT IS: [address]:port
 	if (*str != '[') {
 		config_error_set("Invalid address string '%s'", str);
-		return 0;
+		return (0);
 	}
 
 	const char *ptr = str + 1;
 	const char *x = strrchr(ptr, ']');
 	if (x == NULL) {
 		config_error_set("Invalid address '%s'.", str);
-		return 0;
+		return (0);
 	}
 
+	unsigned addrlen = x - ptr;
 	// address
-	if ((unsigned)(x - ptr) >= sizeof(addr_buf)) {
+	if (addrlen >= sizeof(addr_buf)) {
 		config_error_set("Invalid address '%s'.", str);
-		return 0;
+		return (0);
 	}
-	strncpy(addr_buf, ptr, (x - ptr));
+	strncpy(addr_buf, ptr, addrlen);
 
 	// port
 	if (x[1] != ':' || x[2] == '\0') {
 		config_error_set("Invalid port specifier in string '%s'.", str);
-		return 0;
+		return (0);
 	}
 	ptr = x + 2;
 	x = strchr(ptr, '+');
@@ -366,24 +379,26 @@ config_param_host_port_wildcard(const char *str, char **addr,
 		memcpy(port_buf, ptr, (x - ptr));
 
 	// cert
-	if (cert && x) {
+	if (cert && x)
 		cert_ptr = x + 1;
-	}
 
 	// printf("PARSED ADDR '%s', PORT '%s'\n", addr_buf, port_buf);
 
 	int p = atoi(port_buf);
-	if (p < 1 || p > 65536) {
+	if (p < 0 || p > 65536) {
 		config_error_set("Invalid port number '%s'", port_buf);
-		return 0;
+		return (0);
 	}
 
 	if (strcmp(addr_buf, "*") == 0) {
-		if (wildcard_okay)
+		if (wildcard_okay) {
 			free(*addr);
+			*addr = NULL;
+		}
 		else {
-			config_error_set("Invalid address: wildcards are not allowed.");
-			return 0;
+			config_error_set(
+			    "Invalid address: wildcards are not allowed.");
+			return (0);
 		}
 	} else {
 		*addr = strdup(addr_buf);
@@ -395,48 +410,71 @@ config_param_host_port_wildcard(const char *str, char **addr,
 	/* printf("ADDR FINAL: '%s', '%s', '%s'\n", *addr, *port, */
 	/*     cert ? *cert : ""); */
 
-	return 1;
+	return (1);
 }
 
-int
+static int
 config_param_host_port(char *str, char **addr, char **port)
 {
-	return config_param_host_port_wildcard(str, addr, port, NULL, 0);
+	return (config_param_host_port_wildcard(str, addr, port, NULL, 0));
 }
 
 
-int
-config_param_val_int(char *str, int *dst, int positive_only)
+static int
+config_param_val_int(char *str, int *dst, int non_negative)
 {
-	int num;
+	long  lval;
+	char *ep;
 
 	assert(str != NULL);
-	num = atoi(str);
 
-	if (positive_only && num < 0) {
-		config_error_set("Not a positive number.");
-		return 0;
+	errno = 0;
+	lval = strtol(str, &ep, 10);
+
+	if (*str == '\0' || *ep != '\0') {
+		config_error_set("Not a number.");
+		return (0);
+	}
+	if ((errno == ERANGE && (lval == LONG_MIN || lval == LONG_MAX)) ||
+	    lval < INT_MIN || lval > INT_MAX) {
+		config_error_set("Number out of range.");
+		return (0);
+	}
+	if (non_negative && lval < 0) {
+		config_error_set("Negative number.");
+		return (0);
 	}
 
-	*dst = num;
-	return 1;
+	*dst = (int)lval;
+	return (1);
 }
 
-int
-config_param_val_long(char *str, long *dst, int positive_only)
+static int
+config_param_val_long(char *str, long *dst, int non_negative)
 {
-	long num;
+	long  lval;
+	char *ep;
+
 	assert(str != NULL);
 
-	num = atol(str);
+	errno = 0;
+	lval = strtol(str, &ep, 10);
 
-	if (positive_only && num <= 0) {
-		config_error_set("Not a positive number.");
-		return 0;
+	if (*str == '\0' || *ep != '\0') {
+		config_error_set("Not a number.");
+		return (0);
+	}
+	if (errno == ERANGE && (lval == LONG_MIN || lval == LONG_MAX)) {
+		config_error_set("Number out of range.");
+		return (0);
+	}
+	if (non_negative && lval < 0) {
+		config_error_set("Negative number.");
+		return (0);
 	}
 
-	*dst = num;
-	return 1;
+	*dst = lval;
+	return (1);
 }
 
 static double
@@ -479,9 +517,9 @@ int
 cfg_cert_vfy(struct cfg_cert_file *cf)
 {
 	struct stat st;
+	double d;
 
 	CHECK_OBJ_NOTNULL(cf, CFG_CERT_FILE_MAGIC);
-
 	AN(cf->filename);
 
 	if (cf->filename == NULL || strlen(cf->filename) <= 0)
@@ -515,6 +553,24 @@ cfg_cert_vfy(struct cfg_cert_file *cf)
 		cf->ocsp_mtim = mtim2double(&st);
 	}
 
+	if (cf->priv_key_filename != NULL &&
+	    strlen(cf->priv_key_filename) > 0) {
+
+		if (stat(cf->priv_key_filename, &st) != 0) {
+			config_error_set("Unable to stat private keyfile "
+			    "'%s': %s", cf->priv_key_filename,
+			    strerror(errno));
+			return (0);
+		}
+		if (!S_ISREG(st.st_mode)) {
+			config_error_set("Invalid private keyfile "
+			    "'%s': Not a file.", cf->priv_key_filename);
+			return (0);
+		}
+
+		d = mtim2double(&st);
+		cf->mtim = cf->mtim > d ? cf->mtim : d;
+	}
 	return (1);
 }
 
@@ -524,14 +580,12 @@ cfg_cert_add(struct cfg_cert_file *cf, struct cfg_cert_file **dst)
 	CHECK_OBJ_NOTNULL(cf, CFG_CERT_FILE_MAGIC);
 	AN(dst);
 	CHECK_OBJ_ORNULL(*dst, CFG_CERT_FILE_MAGIC);
-
-	HASH_ADD_KEYPTR(hh, *dst, cf->filename,
-	    strlen(cf->filename), cf);
+	HASH_ADD_KEYPTR(hh, *dst, cf->filename, strlen(cf->filename), cf);
 }
 
 #ifdef USE_SHARED_CACHE
 /* Parse mcast and ttl options */
-int
+static int
 config_param_shcupd_mcastif(char *str, char **iface, char **ttl)
 {
 	char buf[150];
@@ -539,7 +593,7 @@ config_param_shcupd_mcastif(char *str, char **iface, char **ttl)
 
 	if (strlen(str) >= sizeof buf) {
 		config_error_set("Invalid option for IFACE[,TTL]");
-		return 0;
+		return (0);
 	}
 
 	sp = strchr(str, ',');
@@ -549,26 +603,25 @@ config_param_shcupd_mcastif(char *str, char **iface, char **ttl)
 		else
 			*iface = strdup(str);
 		*ttl = NULL;
-		return 1;
+		return (1);
 	}
-	else if (!strncmp(str, "*", sp - str)) {
+	else if (!strncmp(str, "*", sp - str))
 		*iface = NULL;
-	}
 	else {
 		*sp = 0;
 		*iface = strdup(str);
 	}
 	*ttl = strdup(sp + 1);
 
-	return 1;
+	return (1);
 }
 
-int
+static int
 config_param_shcupd_peer(char *str, hitch_config *cfg)
 {
 	if (cfg == NULL) {
 		config_error_set("Configuration pointer is NULL.");
-		return 0;
+		return (0);
 	}
 
 	// parse result
@@ -589,7 +642,7 @@ config_param_shcupd_peer(char *str, hitch_config *cfg)
 		    "Reached maximum number of shared cache update peers (%d).",
 		    MAX_SHCUPD_PEERS
 		);
-		return 0;
+		return (0);
 	}
 
 	// create place for new peer
@@ -623,18 +676,34 @@ config_param_shcupd_peer(char *str, hitch_config *cfg)
 	outta_parse_peer:
 
 	if (! r) {
-		if (addr != NULL) free(addr);
-		if (port != NULL) free(port);
+		free(addr);
+		free(port);
 	} else {
 		cfg->SHCUPD_PEERS[offset].ip = addr;
 		cfg->SHCUPD_PEERS[offset].port = port;
 	}
 
-	return r;
+	return (r);
 }
 
 #endif /* USE_SHARED_CACHE */
 
+static int
+check_frontend_uniqueness(struct front_arg *cur_fa, hitch_config *cfg)
+{
+	struct front_arg *fa, *fatmp;
+	HASH_ITER(hh, cfg->LISTEN_ARGS, fa, fatmp) {
+		if ((cur_fa->ip == NULL || strcmp(cur_fa->ip, fa->ip) == 0) &&
+		    strcmp(cur_fa->port, fa->port) == 0) {
+			config_error_set("Redundant frontend "
+			"(matching IP and port) definition: '%s:%s'.",
+			    fa->ip, fa->port);
+			return (0);
+		}
+	}
+	return(1);
+}
+
 int
 front_arg_add(hitch_config *cfg, struct front_arg *fa)
 {
@@ -665,6 +734,9 @@ front_arg_add(hitch_config *cfg, struct front_arg *fa)
 		return (0);
 	}
 
+	if (check_frontend_uniqueness(fa, cfg) == 0)
+		return (0);
+
 	HASH_ADD_KEYPTR(hh, cfg->LISTEN_ARGS, fa->pspec,
 	    strlen(fa->pspec), fa);
 
@@ -744,6 +816,8 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 		r = config_param_val_int(v, &cfg->BACKLOG, 0);
 	} else if (strcmp(k, CFG_KEEPALIVE) == 0) {
 		r = config_param_val_int(v, &cfg->TCP_KEEPALIVE_TIME, 1);
+	} else if (strcmp(k, CFG_BACKEND_REFRESH) == 0) {
+		r = config_param_val_int(v, &cfg->BACKEND_REFRESH_TIME, 1);
 	}
 #ifdef USE_SHARED_CACHE
 	else if (strcmp(k, CFG_SHARED_CACHE) == 0) {
@@ -801,6 +875,10 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 		}
 	} else if (strcmp(k, CFG_QUIET) == 0) {
 		r = config_param_val_bool(v, &cfg->QUIET);
+	} else if (strcmp(k, CFG_LOG_FILENAME) == 0) {
+		if (strlen(v) > 0) {
+			config_assign_str(&cfg->LOG_FILENAME, v);
+		}
 	} else if (strcmp(k, CFG_SYSLOG) == 0) {
 		r = config_param_val_bool(v, &cfg->SYSLOG);
 	} else if (strcmp(k, CFG_SYSLOG_FACILITY) == 0) {
@@ -891,7 +969,7 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 	return (0);
 }
 
-int
+static int
 config_file_parse(char *file, hitch_config *cfg)
 {
 	FILE *fp = NULL;
@@ -923,61 +1001,64 @@ config_file_parse(char *file, hitch_config *cfg)
 	return (r);
 }
 
-char *
+static char *
 config_disp_str(char *str)
 {
-	return (str == NULL) ? "" : str;
+	return ((str == NULL) ? "" : str);
 }
 
-char *
+static char *
 config_disp_bool(int v)
 {
-	return (v > 0) ? CFG_BOOL_ON : "off";
+	return ((v > 0) ? CFG_BOOL_ON : "off");
 }
 
-char *
+static char *
 config_disp_uid(uid_t uid)
 {
 	memset(tmp_buf, '\0', sizeof(tmp_buf));
-	if (uid == 0 && geteuid() != 0) return tmp_buf;
+	if (uid == 0 && geteuid() != 0)
+		return (tmp_buf);
 	struct passwd *pw = getpwuid(uid);
 	if (pw) {
-		memcpy(tmp_buf, pw->pw_name, strlen(pw->pw_name));
+		strncpy(tmp_buf, pw->pw_name, sizeof(tmp_buf));
+		tmp_buf[sizeof(tmp_buf) - 1] = '\0';
 	}
-	return tmp_buf;
+	return (tmp_buf);
 }
 
-char *
+static char *
 config_disp_gid (gid_t gid)
 {
 	memset(tmp_buf, '\0', sizeof(tmp_buf));
-	if (gid == 0 && geteuid() != 0) return tmp_buf;
+	if (gid == 0 && geteuid() != 0)
+		return (tmp_buf);
 	struct group *gr = getgrgid(gid);
 	if (gr) {
-		memcpy(tmp_buf, gr->gr_name, strlen(gr->gr_name));
+		strncpy(tmp_buf, gr->gr_name, sizeof(tmp_buf));
+		tmp_buf[sizeof(tmp_buf) - 1] = '\0';
 	}
-	return tmp_buf;
+	return (tmp_buf);
 }
 
-char *
+static const char *
 config_disp_hostport(char *host, char *port)
 {
 	memset(tmp_buf, '\0', sizeof(tmp_buf));
 	if (host == NULL && port == NULL)
-		return "";
+		return ("");
 
 	strcat(tmp_buf, "[");
 	if (host == NULL)
 		strcat(tmp_buf, "*");
-	else {
+	else
 		strncat(tmp_buf, host, 40);
-	}
 	strcat(tmp_buf, "]:");
 	strncat(tmp_buf, port, 5);
-	return tmp_buf;
+	return (tmp_buf);
 }
 
-const char *
+static const char *
 config_disp_log_facility (int facility)
 {
 	switch (facility)
@@ -988,7 +1069,7 @@ config_disp_log_facility (int facility)
 #include "sysl_tbl.h"
 #undef SYSLOG_FAC
 		default:
-			return "UNKNOWN";
+			return ("UNKNOWN");
 	}
 }
 
@@ -1018,17 +1099,19 @@ config_print_usage_fd(char *prog, FILE *out)
 	fprintf(out, "\n");
 	fprintf(out, "SOCKET:\n");
 	fprintf(out, "\n");
-	fprintf(out, "  --client                      Enable client proxy mode\n");
-	fprintf(out, "  -b  --backend=[HOST]:PORT     Backend [connect] (default is \"%s\")\n", config_disp_hostport(cfg->BACK_IP, cfg->BACK_PORT));
+	fprintf(out, "      --client                Enable client proxy mode\n");
+	fprintf(out, "  -b  --backend=[HOST]:PORT   Backend [connect] (default is \"%s\")\n", config_disp_hostport(cfg->BACK_IP, cfg->BACK_PORT));
 	fprintf(out, "  -f  --frontend=[HOST]:PORT[+CERT]    Frontend [bind] (default is \"%s\")\n", config_disp_hostport(cfg->LISTEN_DEFAULT->ip, cfg->LISTEN_DEFAULT->port));
-	fprintf(out, "                                (Note: brackets are mandatory in endpoint specifiers.)");
+	fprintf(out, "                                (Note: brackets are mandatory in endpoint specifiers.)\n");
+	fprintf(out, "      --recv-bufsize=SIZE    Receive buffer size on client socket (Default: %d)\n", cfg->RECV_BUFSIZE);
+	fprintf(out, "      --send-bufsize=SIZE    Send buffer size on client socket (Default: %d)\n", cfg->SEND_BUFSIZE);
 
 #ifdef USE_SHARED_CACHE
 	fprintf(out, "\n");
-	fprintf(out, "  -U  --shared-cache-listen=HOST,PORT\n");
+	fprintf(out, "  -U  --shared-cache-listen=[HOST]:PORT\n");
 	fprintf(out, "                              Accept cache updates on UDP (Default: \"%s\")\n", config_disp_hostport(cfg->SHCUPD_IP, cfg->SHCUPD_PORT));
 	fprintf(out, "                              NOTE: This option requires enabled SSL session cache.\n");
-	fprintf(out, "  -P  --shared-cache-peer=HOST,PORT\n");
+	fprintf(out, "  -P  --shared-cache-peer=[HOST]:PORT\n");
 	fprintf(out, "                              Send cache updates to specified peer\n");
 	fprintf(out, "                              NOTE: This option can be specified multiple times.\n");
 	fprintf(out, "  -M  --shared-cache-if=IFACE[,TTL]\n");
@@ -1041,6 +1124,8 @@ config_print_usage_fd(char *prog, FILE *out)
 	fprintf(out, "  -n  --workers=NUM          Number of worker processes (Default: %ld)\n", cfg->NCORES);
 	fprintf(out, "  -B  --backlog=NUM          Set listen backlog size (Default: %d)\n", cfg->BACKLOG);
 	fprintf(out, "  -k  --keepalive=SECS       TCP keepalive on client socket (Default: %d)\n", cfg->TCP_KEEPALIVE_TIME);
+	fprintf(out, "  -R  --backend-refresh=SECS Periodic backend IP lookup, 0 to disable (Default: %d)\n", cfg->BACKEND_REFRESH_TIME);
+
 
 #ifdef USE_SHARED_CACHE
 	fprintf(out, "  -C  --session-cache=NUM    Enable and set SSL session cache to specified number\n");
@@ -1056,8 +1141,9 @@ config_print_usage_fd(char *prog, FILE *out)
 	fprintf(out, "\n");
 	fprintf(out, "LOGGING:\n");
 	fprintf(out, "  -q  --quiet                Be quiet; emit only error messages\n");
+	fprintf(out, "  -l  --log-filename=FILE    Send log message to a logfile instead of stderr/stdout\n");
 	fprintf(out, "  -s  --syslog               Send log message to syslog in addition to stderr/stdout\n");
-	fprintf(out, "  --syslog-facility=FACILITY Syslog facility to use (Default: \"%s\")\n", config_disp_log_facility(cfg->SYSLOG_FACILITY));
+	fprintf(out, "      --syslog-facility=FACILITY    Syslog facility to use (Default: \"%s\")\n", config_disp_log_facility(cfg->SYSLOG_FACILITY));
 	fprintf(out, "\n");
 	fprintf(out, "OTHER OPTIONS:\n");
 	fprintf(out, "      --daemon               Fork into background and become a daemon (Default: %s)\n", config_disp_bool(cfg->DAEMONIZE));
@@ -1092,12 +1178,63 @@ config_print_usage_fd(char *prog, FILE *out)
 	config_destroy(cfg);
 }
 
-void
+static void
 config_print_usage(char *prog)
 {
 	config_print_usage_fd(prog, stdout);
 }
 
+static int
+create_alpn_callback_data(hitch_config *cfg, char **error)
+{
+	size_t i = 1, j, l;
+
+	AN(cfg->ALPN_PROTOS);
+	l = strlen(cfg->ALPN_PROTOS);
+	cfg->ALPN_PROTOS_LV = malloc(l + 1);
+	AN(cfg->ALPN_PROTOS_LV);
+
+	// first remove spaces while copying to cfg->ALPN_PROTOS_LV
+	for(j = 0; j < l; j++)
+		if (!isspace(cfg->ALPN_PROTOS[j])) {
+			cfg->ALPN_PROTOS_LV[i] = cfg->ALPN_PROTOS[j];
+			i++;
+		}
+
+	l = i - 1; // same as before iff cfg->ALPN_PROTOS has no spaces
+	i = 0; // location of next "length" byte
+	for(j = 1; j <= l; j++) {
+		if (cfg->ALPN_PROTOS_LV[j] == ',') {
+			if (i + 1 == j) {
+				*error = "alpn-protos has empty proto in list";
+				return (0); // failure
+			}
+			if (j - i > 256) {
+				free(cfg->ALPN_PROTOS_LV);
+				cfg->ALPN_PROTOS_LV = NULL;
+				*error = "alpn protocol too long";
+				return (0);
+			}
+			cfg->ALPN_PROTOS_LV[i] = (unsigned char)(j - i - 1);
+			i = j;
+		}
+	}
+	if (i == j) {
+		// alpn-protos ends with a comma - we let it slide
+		cfg->ALPN_PROTOS_LV_LEN = l;
+	} else {
+		if (j - i > 256) {
+			free(cfg->ALPN_PROTOS_LV);
+			cfg->ALPN_PROTOS_LV = NULL;
+			*error = "alpn protocol too long";
+			return (0);
+		}
+		cfg->ALPN_PROTOS_LV[i] = (unsigned char)(j - i - 1);
+		cfg->ALPN_PROTOS_LV_LEN = l + 1;
+	}
+	return (1); // ok!
+}
+
 int
 config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 {
@@ -1129,12 +1266,16 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 #endif
 		{ CFG_PIDFILE, 1, NULL, 'p' },
 		{ CFG_KEEPALIVE, 1, NULL, 'k' },
+		{ CFG_BACKEND_REFRESH, 1, NULL, 'R' },
 		{ CFG_CHROOT, 1, NULL, 'r' },
 		{ CFG_USER, 1, NULL, 'u' },
 		{ CFG_GROUP, 1, NULL, 'g' },
 		{ CFG_QUIET, 0, NULL, 'q' },
+		{ CFG_LOG_FILENAME, 1, NULL, 'l' },
 		{ CFG_SYSLOG, 0, NULL, 's' },
 		{ CFG_SYSLOG_FACILITY, 1, NULL, CFG_PARAM_SYSLOG_FACILITY },
+		{ CFG_SEND_BUFSIZE, 1, NULL, CFG_PARAM_SEND_BUFSIZE },
+		{ CFG_RECV_BUFSIZE, 1, NULL, CFG_PARAM_RECV_BUFSIZE },
 		{ CFG_DAEMON, 0, &cfg->DAEMONIZE, 1 },
 		{ CFG_WRITE_IP, 0, &cfg->WRITE_IP_OCTET, 1 },
 		{ CFG_WRITE_PROXY_V1, 0, &cfg->WRITE_PROXY_LINE_V1, 1 },
@@ -1149,7 +1290,7 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		{ "help", 0, NULL, 'h' },
 		{ 0, 0, 0, 0 }
 	};
-#define SHORT_OPTS "c:e:Ob:f:n:B:C:U:p:P:M:k:r:u:g:qstVho:"
+#define SHORT_OPTS "c:e:Ob:f:n:B:l:C:U:p:P:M:k:r:u:g:qstVho:R:"
 
 	if (argc == 1) {
 		config_print_usage(argv[0]);
@@ -1166,6 +1307,13 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 			long_options, &option_index);
 		if (c == -1)
 			break;
+		else if (c == '?') {
+			config_error_set("Invalid command line parameters. "
+			    "Run %s --help for instructions.",
+			    basename(argv[0]));
+			*retval = 1;
+			return (1);
+		}
 		else if (c == CFG_PARAM_CFGFILE) {
 			if (config_file_parse(optarg, cfg) != 0) {
 				*retval = 1;
@@ -1192,65 +1340,44 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		case CFG_PARAM_CFGFILE:
 			/* Handled above */
 			break;
-		case CFG_PARAM_SYSLOG_FACILITY:
-			ret = config_param_validate(CFG_SYSLOG_FACILITY, optarg, cfg, NULL, 0);
+#define CFG_ARG(opt, key)							\
+		case opt:							\
+			ret = config_param_validate(key, optarg, cfg, NULL, 0);	\
 			break;
-		case 'c':
-			ret = config_param_validate(CFG_CIPHERS, optarg, cfg, NULL, 0);
-			break;
-		case 'e':
-			ret = config_param_validate(CFG_SSL_ENGINE, optarg, cfg, NULL, 0);
-			break;
-		case 'O':
-			ret = config_param_validate(CFG_PREFER_SERVER_CIPHERS, CFG_BOOL_ON, cfg, NULL, 0);
-			break;
-		case 'b':
-			ret = config_param_validate(CFG_BACKEND, optarg, cfg, NULL, 0);
-			break;
-		case 'f':
-			ret = config_param_validate(CFG_FRONTEND, optarg, cfg, NULL, 0);
-			break;
-		case 'n':
-			ret = config_param_validate(CFG_WORKERS, optarg, cfg, NULL, 0);
-			break;
-		case 'B':
-			ret = config_param_validate(CFG_BACKLOG, optarg, cfg, NULL, 0);
+#define CFG_ON(opt, key)							\
+		case opt:							\
+			ret = config_param_validate(key, CFG_BOOL_ON, cfg, 	\
+			    NULL, 0);						\
 			break;
+CFG_ARG(CFG_PARAM_SYSLOG_FACILITY, CFG_SYSLOG_FACILITY);
+CFG_ARG(CFG_PARAM_SEND_BUFSIZE, CFG_SEND_BUFSIZE);
+CFG_ARG(CFG_PARAM_RECV_BUFSIZE, CFG_RECV_BUFSIZE);
+CFG_ARG(CFG_PARAM_ALPN_PROTOS, CFG_ALPN_PROTOS);
+CFG_ARG('c', CFG_CIPHERS);
+CFG_ARG('e', CFG_SSL_ENGINE);
+CFG_ARG('b', CFG_BACKEND);
+CFG_ARG('f', CFG_FRONTEND);
+CFG_ARG('n', CFG_WORKERS);
+CFG_ARG('B', CFG_BACKLOG);
 #ifdef USE_SHARED_CACHE
-		case 'C':
-			ret = config_param_validate(CFG_SHARED_CACHE, optarg, cfg, NULL, 0);
-			break;
-		case 'U':
-			ret = config_param_validate(CFG_SHARED_CACHE_LISTEN, optarg, cfg, NULL, 0);
-			break;
-		case 'P':
-			ret = config_param_validate(CFG_SHARED_CACHE_PEER, optarg, cfg, NULL, 0);
-			break;
-		case 'M':
-			ret =config_param_validate(CFG_SHARED_CACHE_MCASTIF, optarg, cfg, NULL, 0);
-			break;
+CFG_ARG('C', CFG_SHARED_CACHE);
+CFG_ARG('U', CFG_SHARED_CACHE_LISTEN);
+CFG_ARG('P', CFG_SHARED_CACHE_PEER);
+CFG_ARG('M', CFG_SHARED_CACHE_MCASTIF);
 #endif
-		case 'p':
-			ret = config_param_validate(CFG_PIDFILE, optarg, cfg, NULL, 0);
-			break;
-		case 'k':
-			ret = config_param_validate(CFG_KEEPALIVE, optarg, cfg, NULL, 0);
-			break;
-		case 'r':
-			ret = config_param_validate(CFG_CHROOT, optarg, cfg, NULL, 0);
-			break;
-		case 'u':
-			ret = config_param_validate(CFG_USER, optarg, cfg, NULL, 0);
-			break;
-		case 'g':
-			ret = config_param_validate(CFG_GROUP, optarg, cfg, NULL, 0);
-			break;
-		case 'q':
-			ret = config_param_validate(CFG_QUIET, CFG_BOOL_ON, cfg, NULL, 0);
-			break;
-		case 's':
-			ret = config_param_validate(CFG_SYSLOG, CFG_BOOL_ON, cfg, NULL, 0);
-			break;
+CFG_ARG('p', CFG_PIDFILE);
+CFG_ARG('k', CFG_KEEPALIVE);
+CFG_ARG('R', CFG_BACKEND_REFRESH);
+CFG_ARG('r', CFG_CHROOT);
+CFG_ARG('u', CFG_USER);
+CFG_ARG('g', CFG_GROUP);
+CFG_ARG('o', CFG_OCSP_DIR);
+CFG_ON('O', CFG_PREFER_SERVER_CIPHERS);
+CFG_ON('q', CFG_QUIET);
+CFG_ARG('l', CFG_LOG_FILENAME);
+CFG_ON('s', CFG_SYSLOG);
+#undef CFG_ARG
+#undef CFG_ON
 		case 't':
 			cfg->TEST = 1;
 			break;
@@ -1264,12 +1391,6 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 			*retval = 0;
 			return (1);
 			break;
-		case 'o':
-			ret = config_param_validate(CFG_OCSP_DIR, optarg, cfg, NULL, 0);
-			break;
-		case CFG_PARAM_ALPN_PROTOS:
-			ret = config_param_validate(CFG_ALPN_PROTOS, optarg, cfg, NULL, 0);
-			break;
 
 		default:
 			config_error_set("Invalid command line parameters. "
@@ -1319,7 +1440,6 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 
 
 	if (cfg->DAEMONIZE) {
-		cfg->SYSLOG = 1;
 		cfg->QUIET = 1;
 	}
 
@@ -1405,15 +1525,13 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 
 	if (cfg->PMODE == SSL_SERVER && cfg->CERT_DEFAULT == NULL) {
 		struct front_arg *fa, *fatmp;
-		HASH_ITER(hh, cfg->LISTEN_ARGS, fa, fatmp) {
+		HASH_ITER(hh, cfg->LISTEN_ARGS, fa, fatmp)
 			if (HASH_CNT(hh, fa->certs) == 0) {
 				config_error_set("No x509 certificate PEM file "
 				    "specified for frontend '%s'!", fa->pspec);
 				*retval = 1;
 				return (1);
-
 			}
-		}
 	}
 
 	if (cfg->OCSP_DIR != NULL) {
@@ -1439,53 +1557,3 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 
 	return (0);
 }
-
-int create_alpn_callback_data(hitch_config *cfg, char **error)
-{
-	size_t i = 1, j, l;
-
-	AN(cfg->ALPN_PROTOS);
-	l = strlen(cfg->ALPN_PROTOS);
-	cfg->ALPN_PROTOS_LV = malloc(l + 1);
-	AN(cfg->ALPN_PROTOS_LV);
-
-	// first remove spaces while copying to cfg->ALPN_PROTOS_LV
-	for(j = 0; j < l; j++) {
-		if (!isspace(cfg->ALPN_PROTOS[j])) {
-			cfg->ALPN_PROTOS_LV[i] = cfg->ALPN_PROTOS[j];
-			i++;
-		}
-	}
-	l = i - 1; // same as before iff cfg->ALPN_PROTOS has no spaces
-	i = 0; // location of next "length" byte
-	for(j = 1; j <= l; j++) {
-		if (cfg->ALPN_PROTOS_LV[j] == ',') {
-			if (i + 1 == j) {
-				*error = "alpn-protos has empty proto in list";
-				return (0); // failure
-			}
-			if (j - i > 256) {
-				free(cfg->ALPN_PROTOS_LV);
-				cfg->ALPN_PROTOS_LV = NULL;
-				*error = "alpn protocol too long";
-				return (0);
-			}
-			cfg->ALPN_PROTOS_LV[i] = (unsigned char)(j - i - 1);
-			i = j;
-		}
-	}
-	if (i == j) {
-		// alpn-protos ends with a comma - we let it slide
-		cfg->ALPN_PROTOS_LV_LEN = l;
-	} else {
-		if (j - i > 256) {
-			free(cfg->ALPN_PROTOS_LV);
-			cfg->ALPN_PROTOS_LV = NULL;
-			*error = "alpn protocol too long";
-			return (0);
-		}
-		cfg->ALPN_PROTOS_LV[i] = (unsigned char)(j - i - 1);
-		cfg->ALPN_PROTOS_LV_LEN = l + 1;
-	}
-	return (1); // ok!
-}
