@@ -35,34 +35,42 @@
   *
   */
 
-#include <netdb.h>
+#include "config.h"
+
+#include <sys/socket.h>
+#include <sys/time.h>
+#include <sys/types.h>
+#include <sys/wait.h>  /* WAIT_PID */
+
+#ifdef __linux__
+#include <sys/prctl.h>
+#endif
+
+#ifdef __sun
+#include <sys/filio.h>
+#endif
+
+#include <net/if.h>
 #include <netinet/in.h>
 #include <netinet/tcp.h>  /* TCP_NODELAY */
-#include <net/if.h>
-
-#include <libgen.h>
-#include <sys/socket.h>
-#include <sys/wait.h>  /* WAIT_PID */
 
 #include <ctype.h>
 #include <errno.h>
 #include <getopt.h>
 #include <grp.h>
+#include <libgen.h>
 #include <limits.h>
+#include <netdb.h>
 #include <pwd.h>
 #include <sched.h>
 #include <signal.h>
 #include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h>
 #include <time.h>
 #include <unistd.h>
 
-#ifdef __linux__
-#include <sys/prctl.h>
-#endif
-
-#include "config.h"
 #include "configuration.h"
 #include "hitch.h"
 #include "hssl_locks.h"
@@ -71,6 +79,7 @@
 #include "shctx.h"
 #include "foreign/vpf.h"
 #include "foreign/uthash.h"
+#include "foreign/vsa.h"
 
 #ifndef MSG_NOSIGNAL
 # define MSG_NOSIGNAL 0
@@ -109,7 +118,14 @@ hitch_config *CONFIG;
 /* Worker proc's read side of mgt->worker pipe(2) */
 static ev_io mgt_rd;
 
-static struct addrinfo *backaddr;
+struct backend {
+	unsigned		magic;
+#define BACKEND_MAGIC 0x41c09397
+	struct suckaddr		*backaddr;
+	int 			ref;
+};
+
+static struct backend *backaddr;
 static pid_t master_pid;
 static pid_t ocsp_proc_pid;
 static int core_id;
@@ -231,7 +247,7 @@ union ha_proxy_v2_addr {
 	struct {        /* for AF_UNIX sockets, len = 216 */
 		uint8_t src_addr[108];
 		uint8_t dst_addr[108];
-	} unix;
+	} local;
 };
 
 struct ha_proxy_v2_hdr {
@@ -243,6 +259,20 @@ struct ha_proxy_v2_hdr {
 	union ha_proxy_v2_addr	addr;
 };
 
+enum worker_update_type {
+	WORKER_GEN,
+	BACKEND_REFRESH
+};
+
+union worker_update_payload {
+	unsigned		gen;
+	struct sockaddr_storage	addr;
+};
+
+struct worker_update {
+	enum worker_update_type		type;
+	union worker_update_payload 	payload;
+};
 
 /* set a file descriptor (socket) to non-blocking mode */
 static int
@@ -683,7 +713,7 @@ load_privatekey(SSL_CTX *ctx, const char *file)
 		return NULL;
 	}
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
 #define SSL_CTX_get_default_passwd_cb(ctx) (ctx->default_passwd_callback)
 #define SSL_CTX_get_default_passwd_cb_userdata(ctx) (ctx->default_passwd_callback_userdata)
 #endif
@@ -692,6 +722,11 @@ load_privatekey(SSL_CTX *ctx, const char *file)
 	    SSL_CTX_get_default_passwd_cb_userdata(ctx));
 	BIO_free(bio);
 
+	if (!pkey) {
+		log_ssl_error(NULL, "{core} PEM_read_bio_PrivateKey");
+		return NULL;
+	}
+
 	return (pkey);
 }
 
@@ -861,6 +896,9 @@ make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
 #endif
 #ifdef SSL_OP_SINGLE_DH_USE
 	ssloptions |= SSL_OP_SINGLE_DH_USE;
+#endif
+#ifdef SSL_OP_SINGLE_ECDH_USE
+	ssloptions |= SSL_OP_SINGLE_ECDH_USE;
 #endif
 	if (!(selected_protos & SSLv3_PROTO))
 		ssloptions |= SSL_OP_NO_SSLv3;
@@ -883,7 +921,6 @@ make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
 	if (CONFIG->ALPN_PROTOS != NULL)
 		SSL_CTX_set_next_protos_advertised_cb(ctx, npn_select_cb, NULL);
 #endif
-	AN(SSL_CTX_set_default_verify_paths(ctx));
 
 	if (ciphers != NULL) {
 		if (SSL_CTX_set_cipher_list(ctx, ciphers) != 1) {
@@ -903,6 +940,10 @@ make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
 	sc->staple_vfy = cf->ocsp_vfy;
 	VTAILQ_INIT(&sc->sni_list);
 
+	if (sc->staple_vfy > 0 ||
+	    (sc-> staple_vfy < 0 && CONFIG->OCSP_VFY))
+		AN(SSL_CTX_set_default_verify_paths(ctx));
+
 	if (CONFIG->PMODE == SSL_CLIENT)
 		return (sc);
 
@@ -914,11 +955,23 @@ make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
 		return (NULL);
 	}
 
-	pkey = load_privatekey(ctx, cf->filename);
+	pkey = 0;
+	if (cf->priv_key_filename != 0) {
+		pkey = load_privatekey(ctx, cf->priv_key_filename);
+		if (!pkey) {
+			ERR("Error loading private key (%s)\n",
+			    cf->priv_key_filename);
+			sctx_free(sc, NULL);
+			return (NULL);
+		}
+	}
 	if (!pkey) {
-		ERR("Error loading private key (%s)\n", cf->filename);
-		sctx_free(sc, NULL);
-		return (NULL);
+		pkey = load_privatekey(ctx, cf->filename);
+		if (!pkey) {
+			ERR("Error loading private key (%s)\n", cf->filename);
+			sctx_free(sc, NULL);
+			return (NULL);
+		}
 	}
 
 	if (SSL_CTX_use_PrivateKey(ctx, pkey) <= 0) {
@@ -1404,12 +1457,64 @@ create_frontend(const struct front_arg *fa)
 	return (fr);
 }
 
+static const void *
+Get_Sockaddr(const struct sockaddr *sa, socklen_t *sl);
+
+static struct backend *
+backend_create(struct sockaddr *sa)
+{
+	socklen_t len;
+	const void *addr;
+	struct backend *b;
+	struct suckaddr *su;
+
+	addr = Get_Sockaddr(sa, &len);
+	su = VSA_Malloc(addr, len);
+	ALLOC_OBJ(b, BACKEND_MAGIC);
+	b->backaddr = su;
+	b->ref = 1;
+	return (b);
+}
+
+static struct backend *
+backend_ref(void)
+{
+	CHECK_OBJ_NOTNULL(backaddr, BACKEND_MAGIC);
+	AN(backaddr->ref);
+	backaddr->ref++;
+	return (backaddr);
+}
+
+
+void
+backend_deref(struct backend **be)
+{
+	struct backend *b;
+	b = *be;
+
+	CHECK_OBJ_NOTNULL(b, BACKEND_MAGIC);
+	AN(b->ref);
+	b->ref--;
+
+	if (b->ref == 0) {
+		free(b->backaddr);
+		FREE_OBJ(*be);
+	}
+}
+
 /* Initiate a clear-text nonblocking connect() to the backend IP on behalf
  * of a newly connected upstream (encrypted) client */
 static int
-create_back_socket()
+create_back_socket(struct backend *b)
 {
-	int s = socket(backaddr->ai_family, SOCK_STREAM, IPPROTO_TCP);
+	socklen_t len;
+	const void *addr;
+
+	CHECK_OBJ_NOTNULL(b, BACKEND_MAGIC);
+	addr = VSA_Get_Sockaddr(b->backaddr, &len);
+	AN(addr);
+	int s = socket(((const struct sockaddr *)addr)->sa_family,
+		       SOCK_STREAM, IPPROTO_TCP);
 
 	if (s == -1)
 		return -1;
@@ -1473,6 +1578,7 @@ shutdown_proxy(proxystate *ps, SHUTDOWN_REQUESTOR req)
 
 		close(ps->fd_up);
 		close(ps->fd_down);
+		backend_deref(&ps->backend);
 
 		ringbuffer_cleanup(&ps->ring_clear2ssl);
 		ringbuffer_cleanup(&ps->ring_ssl2clear);
@@ -1512,16 +1618,25 @@ static int
 start_connect(proxystate *ps)
 {
 	int t = 1;
+	socklen_t len;
+	const void *addr;
+
 	CHECK_OBJ_NOTNULL(ps, PROXYSTATE_MAGIC);
-	t = connect(ps->fd_down, backaddr->ai_addr, backaddr->ai_addrlen);
+	CHECK_OBJ_NOTNULL(ps->backend, BACKEND_MAGIC);
+	addr = VSA_Get_Sockaddr(ps->backend->backaddr, &len);
+	AN(addr);
+
+	t = connect(ps->fd_down, addr, len);
 	if (t == 0 || errno == EINPROGRESS || errno == EINTR) {
 		ev_io_start(loop, &ps->ev_w_connect);
 		ev_timer_start(loop, &ps->ev_t_connect);
-		return 0;
+		return (0);
 	}
+
 	ERR("{backend-connect}: %s\n", strerror(errno));
 	shutdown_proxy(ps, SHUTDOWN_HARD);
-	return -1;
+
+	return (-1);
 }
 
 /* Read some data from the backend when libev says data is available--
@@ -1608,11 +1723,17 @@ handle_connect(struct ev_loop *loop, ev_io *w, int revents)
 {
 	int t, r;
 	proxystate *ps;
+	socklen_t len;
+	const void *addr;
 
 	(void)revents;
 	CAST_OBJ_NOTNULL(ps, w->data, PROXYSTATE_MAGIC);
+	CHECK_OBJ_NOTNULL(ps->backend, BACKEND_MAGIC);
+	addr = VSA_Get_Sockaddr(ps->backend->backaddr, &len);
+	AN(addr);
+
+	t = connect(ps->fd_down, addr, len);
 
-	t = connect(ps->fd_down, backaddr->ai_addr, backaddr->ai_addrlen);
 	if (!t || errno == EISCONN || !errno) {
 		ev_io_stop(loop, &ps->ev_w_connect);
 		ev_timer_stop(loop, &ps->ev_t_connect);
@@ -1920,6 +2041,7 @@ static void end_handshake(proxystate *ps) {
 		} else if (CONFIG->WRITE_IP_OCTET) {
 			write_ip_octet(ps);
 		}
+
 		/* start connect now */
 		if (0 != start_connect(ps))
 			return;
@@ -2262,13 +2384,24 @@ handle_accept(struct ev_loop *loop, ev_io *w, int revents)
 
 	settcpkeepalive(client);
 
-	int back = create_back_socket();
-	if (back == -1) {
+	ALLOC_OBJ(ps, PROXYSTATE_MAGIC);
+	if (ps == NULL) {
+		(void)close(client);
+		ERR("{malloc-err}: %s\n", strerror(errno));
+		return;
+	}
+
+	ps->backend = backend_ref();
+	ps->fd_down = create_back_socket(ps->backend);
+	if (ps->fd_down == -1) {
 		(void) close(client);
+		backend_deref(&ps->backend);
+		free(ps);
 		ERR("{backend-socket}: %s\n", strerror(errno));
 		return;
 	}
 
+
 	CAST_OBJ_NOTNULL(fr, w->data, FRONTEND_MAGIC);
 	if (fr->ssl_ctxs != NULL)
 		CAST_OBJ_NOTNULL(so, fr->ssl_ctxs, SSLCTX_MAGIC);
@@ -2277,21 +2410,14 @@ handle_accept(struct ev_loop *loop, ev_io *w, int revents)
 
 	SSL *ssl = SSL_new(so->ctx);
 	if (ssl == NULL) {
-		(void)close(back);
+		(void)close(ps->fd_down);
 		(void)close(client);
+		backend_deref(&ps->backend);
+		free(ps);
 		ERR("{SSL_new}: %s\n", strerror(errno));
 		return;
 	}
 
-	ALLOC_OBJ(ps, PROXYSTATE_MAGIC);
-	if (ps == NULL) {
-		SSL_free(ssl);
-		(void)close(back);
-		(void)close(client);
-		ERR("{malloc-err}: %s\n", strerror(errno));
-		return;
-	}
-
 	long mode = SSL_MODE_ENABLE_PARTIAL_WRITE;
 #ifdef SSL_MODE_RELEASE_BUFFERS
 	mode |= SSL_MODE_RELEASE_BUFFERS;
@@ -2301,7 +2427,6 @@ handle_accept(struct ev_loop *loop, ev_io *w, int revents)
 	SSL_set_fd(ssl, client);
 
 	ps->fd_up = client;
-	ps->fd_down = back;
 	ps->ssl = ssl;
 	ps->want_shutdown = 0;
 	ps->clear_connected = 0;
@@ -2325,13 +2450,12 @@ handle_accept(struct ev_loop *loop, ev_io *w, int revents)
 	    CONFIG->SSL_HANDSHAKE_TIMEOUT, 0.);
 
 	ev_io_init(&ps->ev_proxy, client_proxy_proxy, client, EV_READ);
-
-	ev_io_init(&ps->ev_w_connect, handle_connect, back, EV_WRITE);
+	ev_io_init(&ps->ev_w_connect, handle_connect, ps->fd_down, EV_WRITE);
 	ev_timer_init(&ps->ev_t_connect, connect_timeout,
 	    CONFIG->BACKEND_CONNECT_TIMEOUT, 0.);
 
-	ev_io_init(&ps->ev_w_clear, clear_write, back, EV_WRITE);
-	ev_io_init(&ps->ev_r_clear, clear_read, back, EV_READ);
+	ev_io_init(&ps->ev_w_clear, clear_write, ps->fd_down, EV_WRITE);
+	ev_io_init(&ps->ev_r_clear, clear_read, ps->fd_down, EV_READ);
 
 	ps->ev_r_ssl.data = ps;
 	ps->ev_w_ssl.data = ps;
@@ -2381,16 +2505,36 @@ check_ppid(struct ev_loop *loop, ev_timer *w, int revents)
 	}
 }
 
+static const void *
+Get_Sockaddr(const struct sockaddr *sa, socklen_t *sl)
+{
+	AN(sa);
+	AN(sl);
+
+	switch (sa->sa_family) {
+	case PF_INET:
+		*sl = sizeof(struct sockaddr_in);
+		break;
+	case PF_INET6:
+		*sl = sizeof(struct sockaddr_in6);
+		break;
+	default:
+		*sl = 0;
+		return (NULL);
+	}
+	return sa;
+}
+
 static void
 handle_mgt_rd(struct ev_loop *loop, ev_io *w, int revents)
 {
-	unsigned cg;
 	ssize_t r;
 	struct frontend *fr;
 	struct listen_sock *ls;
+	struct worker_update wu;
 
 	(void) revents;
-	r = read(w->fd, &cg, sizeof(cg));
+	r = read(w->fd, &wu, sizeof(wu));
 	if (r  == -1) {
 		if (errno == EWOULDBLOCK || errno == EAGAIN)
 			return;
@@ -2406,7 +2550,7 @@ handle_mgt_rd(struct ev_loop *loop, ev_io *w, int revents)
 		_exit(1);
 	}
 
-	if (cg != worker_gen) {
+	if (wu.type == WORKER_GEN && wu.payload.gen != worker_gen) {
 		/* This means this process has reached its retirement age. */
 		worker_state = WORKER_EXITING;
 
@@ -2421,10 +2565,18 @@ handle_mgt_rd(struct ev_loop *loop, ev_io *w, int revents)
 		}
 
 		check_exit_state();
-	}
 
-	LOGL("Worker %d (gen: %d): State %s\n", core_id, worker_gen,
-	    (worker_state == WORKER_EXITING) ? "EXITING" : "ACTIVE");
+		LOGL("Worker %d (gen: %d): State %s\n", core_id, worker_gen,
+		(worker_state == WORKER_EXITING) ? "EXITING" : "ACTIVE");
+	}
+	else if (wu.type == BACKEND_REFRESH) {
+		struct backend *b;
+		b = backend_create((struct sockaddr *)&wu.payload.addr);
+		backend_deref(&backaddr);
+		backaddr = b;
+		AN(VSA_Sane(backaddr->backaddr));
+	} else
+		WRONG("Invalid worker update state");
 }
 
 static void
@@ -2484,9 +2636,13 @@ handle_clear_accept(struct ev_loop *loop, ev_io *w, int revents)
 
 	settcpkeepalive(client);
 
-	int back = create_back_socket();
-	if (back == -1) {
+	ALLOC_OBJ(ps, PROXYSTATE_MAGIC);
+	ps->backend = backend_ref();
+	ps->fd_down = create_back_socket(ps->backend);
+	if (ps->fd_down == -1) {
+		backend_deref(&ps->backend);
 		close(client);
+		free(ps);
 		ERR("{backend-socket}: %s\n", strerror(errno));
 		return;
 	}
@@ -2503,14 +2659,11 @@ handle_clear_accept(struct ev_loop *loop, ev_io *w, int revents)
 #endif
 	SSL_set_mode(ssl, mode);
 	SSL_set_connect_state(ssl);
-	SSL_set_fd(ssl, back);
+	SSL_set_fd(ssl, ps->fd_down);
 	if (client_session)
 		SSL_set_session(ssl, client_session);
 
-	ALLOC_OBJ(ps, PROXYSTATE_MAGIC);
-
 	ps->fd_up = client;
-	ps->fd_down = back;
 	ps->ssl = ssl;
 	ps->want_shutdown = 0;
 	ps->clear_connected = 1;
@@ -2526,17 +2679,19 @@ handle_clear_accept(struct ev_loop *loop, ev_io *w, int revents)
 	ev_io_init(&ps->ev_r_clear, clear_read, client, EV_READ);
 	ev_io_init(&ps->ev_w_clear, clear_write, client, EV_WRITE);
 
-	ev_io_init(&ps->ev_w_connect, handle_connect, back, EV_WRITE);
+	ev_io_init(&ps->ev_w_connect, handle_connect, ps->fd_down, EV_WRITE);
 	ev_timer_init(&ps->ev_t_connect, connect_timeout,
 	    CONFIG->BACKEND_CONNECT_TIMEOUT, 0.);
 
-	ev_io_init(&ps->ev_r_handshake, client_handshake, back, EV_READ);
-	ev_io_init(&ps->ev_w_handshake, client_handshake, back, EV_WRITE);
+	ev_io_init(&ps->ev_r_handshake, client_handshake,
+	    ps->fd_down, EV_READ);
+	ev_io_init(&ps->ev_w_handshake, client_handshake,
+	    ps->fd_down, EV_WRITE);
 	ev_timer_init(&ps->ev_t_handshake, handshake_timeout,
 	    CONFIG->SSL_HANDSHAKE_TIMEOUT, 0.);
 
-	ev_io_init(&ps->ev_w_ssl, ssl_write, back, EV_WRITE);
-	ev_io_init(&ps->ev_r_ssl, ssl_read, back, EV_READ);
+	ev_io_init(&ps->ev_w_ssl, ssl_write, ps->fd_down, EV_WRITE);
+	ev_io_init(&ps->ev_r_ssl, ssl_read, ps->fd_down, EV_READ);
 
 	ps->ev_r_ssl.data = ps;
 	ps->ev_w_ssl.data = ps;
@@ -2715,26 +2870,55 @@ drop_privileges(void)
 #endif
 }
 
-void
-init_globals(void)
-{
-	/* backaddr */
+static int
+get_backend_addrinfo(void) {
+	struct addrinfo *result;
 	struct addrinfo hints;
-
-	VTAILQ_INIT(&frontends);
-	VTAILQ_INIT(&worker_procs);
+	struct backend *b;
 
 	memset(&hints, 0, sizeof hints);
 	hints.ai_family = AF_UNSPEC;
 	hints.ai_socktype = SOCK_STREAM;
 	hints.ai_flags = 0;
 	const int gai_err = getaddrinfo(CONFIG->BACK_IP, CONFIG->BACK_PORT,
-	    &hints, &backaddr);
+	    &hints, &result);
 	if (gai_err != 0) {
 		ERR("{getaddrinfo-backend}: %s\n", gai_strerror(gai_err));
 		exit(1);
 	}
 
+	b = backend_create(result->ai_addr);
+	freeaddrinfo(result);
+
+	if (backaddr == NULL) {
+		backaddr = b;
+		return (1);
+	}
+
+	if (VSA_Compare(backaddr->backaddr, b->backaddr) != 0) {
+		free(backaddr);
+		backaddr = b;
+		return (1);
+	}
+
+	backend_deref(&b);
+	AZ(b);
+
+	return (0);
+}
+
+void
+init_globals(void)
+{
+	struct addrinfo hints;
+
+	VTAILQ_INIT(&frontends);
+	VTAILQ_INIT(&worker_procs);
+
+	get_backend_addrinfo();
+
+	(void)hints;
+
 #ifdef USE_SHARED_CACHE
 	if (CONFIG->SHARED_CACHE) {
 		/* cache update peers addresses */
@@ -2791,6 +2975,8 @@ start_workers(int start_index, int count)
 		} else if (c->pid == 0) { /* child */
 			close(pfd[1]);
 			FREE_OBJ(c);
+			if (CONFIG->CHROOT && CONFIG->CHROOT[0])
+				change_root();
 			if (CONFIG->UID >= 0 || CONFIG->GID >= 0)
 				drop_privileges();
 			if (geteuid() == 0) {
@@ -3401,15 +3587,50 @@ cert_query(hitch_config *cfg, struct cfg_tpc_obj_head *cfg_objs)
 }
 
 static void
-reconfigure(int argc, char **argv)
+notify_workers(struct worker_update *wu)
 {
 	struct worker_proc *c;
+	int i;
+	VTAILQ_FOREACH(c, &worker_procs, list) {
+		if ((wu->type == WORKER_GEN && wu->payload.gen != c->gen) ||
+		     (wu->type == BACKEND_REFRESH)) {
+			errno = 0;
+			do {
+				i = write(c->pfd, (void*)wu, sizeof(*wu));
+				if (i == -1 && errno != EINTR) {
+					if (wu->type == WORKER_GEN)
+						ERR("WARNING: {core} Unable to "
+						"gracefully reload worker %d"
+						" (%s).\n",
+						c->pid, strerror(errno));
+					else
+						ERR("WARNING: {core} Unable to "
+						"notify worker %d "
+						"with changed backend address (%s).\n",
+						c->pid, strerror(errno));
+
+					(void)kill(c->pid, SIGTERM);
+					break;
+				}
+			} while (i == -1 && errno == EINTR);
+
+			if (wu->type == WORKER_GEN)
+				(void)close(c->pfd);
+
+		}
+	}
+}
+
+static void
+reconfigure(int argc, char **argv)
+{
 	hitch_config *cfg_new;
-	int i, rv;
+	int rv;
 	struct cfg_tpc_obj_head cfg_objs;
 	struct cfg_tpc_obj *cto, *cto_tmp;
 	struct timeval tv;
 	double t0, t1;
+	struct worker_update wu;
 
 	LOGL("Received SIGHUP: Initiating configuration reload.\n");
 	AZ(gettimeofday(&tv, NULL));
@@ -3453,33 +3674,50 @@ reconfigure(int argc, char **argv)
 
 	worker_gen++;
 	start_workers(0, CONFIG->NCORES);
-	VTAILQ_FOREACH(c, &worker_procs, list) {
-		if (c->gen != worker_gen) {
-			errno = 0;
-			do {
-				i = write(c->pfd, &worker_gen,
-				    sizeof(worker_gen));
-				if (i == -1 && errno != EINTR) {
-					ERR("WARNING: {core} Unable to "
-					    "gracefully reload worker %d"
-					    " (%s).\n",
-					    c->pid, strerror(errno));
-					(void)kill(c->pid, SIGTERM);
-					break;
-				}
-			} while (i == -1 && errno == EINTR);
-		}
-	}
+
+	wu.type = WORKER_GEN;
+	wu.payload.gen = worker_gen;
+	notify_workers(&wu);
 
 	if (CONFIG->OCSP_DIR != NULL) {
 		(void) kill(ocsp_proc_pid, SIGTERM);
-		start_ocsp_proc();
+		/*
+		 * Restarting the OCSP process is taken
+		 * care of in do_wait
+		 */
 	}
 
 	config_destroy(CONFIG);
 	CONFIG = cfg_new;
 }
 
+void
+sleep_and_refresh(hitch_config *CONFIG)
+{
+	/* static backend address */
+	if (!CONFIG->BACKEND_REFRESH_TIME) {
+		pause();
+		return;
+	}
+
+	int rv = 0;
+	while (1) {
+		rv = usleep(CONFIG->BACKEND_REFRESH_TIME*1000000);
+		if (rv == -1 && errno == EINTR)
+			break;
+		else if(get_backend_addrinfo()) {
+			struct worker_update wu;
+			wu.type = BACKEND_REFRESH;
+			socklen_t len;
+			const void *addr =
+			    VSA_Get_Sockaddr(backaddr->backaddr, &len);
+			AN(addr);
+			memcpy(&(wu.payload.addr), addr, len);
+			notify_workers(&wu);
+		}
+	}
+}
+
 /* Process command line args, create the bound socket,
  * spawn child (worker) processes, and respawn if any die */
 int
@@ -3560,8 +3798,11 @@ main(int argc, char **argv)
 	}
 #endif /* USE_SHARED_CACHE */
 
-	if (CONFIG->CHROOT && CONFIG->CHROOT[0])
-		change_root();
+	if (CONFIG->CHROOT && CONFIG->CHROOT[0] && geteuid() != 0) {
+		ERR("{core} ERROR: chroot requires hitch to be"
+		    " started as root.\n");
+		exit(1);
+	}
 
 	if (geteuid() == 0 && CONFIG->UID < 0) {
 		ERR("{core} ERROR: Refusing to run workers as root.\n");
@@ -3609,8 +3850,10 @@ main(int argc, char **argv)
 				ev_loop(loop, EVRUN_ONCE);
 			}
 		} else
+			sleep_and_refresh(CONFIG);
 #else
-			pause();
+
+		sleep_and_refresh(CONFIG);
 		/* Sleep and let the children work.
 		 * Parent will be woken up if a signal arrives */
 #endif /* USE_SHARED_CACHE */
